﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace ActiveXTestVCS
{
    public partial class frmMain : Form
    {
        protected short m_iPortID;
        protected int m_iSelectedTableItem;
        protected int m_iPreSelectedTableItem;

        public frmMain()
        {
            InitializeComponent();
        }

        private void btnExit_Click(object sender, EventArgs e)
        {
            if (axTrioPC1.IsOpen(0))
                axTrioPC1.Close(0);
            Close();
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            if (axTrioPC1.IsOpen(0))
            {
                btnClose.Enabled = false;
                axTrioPC1.Close(0);
                btnOpen.Enabled = true;
                btnRefresh.Enabled = false;
                axTrioPC1.Refresh();
            }
        }

        private void btnOpen_Click(object sender, EventArgs e)
        {
            short iPortID = -1,
                iPortType = -1;

            btnOpen.Enabled = false;
            btnRefresh.Enabled = false;
            switch (cbConnectionType.Text)
            {
            case "USB":
                iPortType = 0;
                iPortID = 0;
                break;
            case "Ethernet":
                axTrioPC1.HostAddress = tbHostAddress.Text;
                iPortType = 2;
                iPortID = 3240;
                break;
            case "PCI":
                axTrioPC1.Board = Convert.ToInt32(tbChannel.Text);
                iPortType = 3;
                iPortID = 0;
                break;
            }

            if ((iPortType != -1) && (iPortID != -1))
                axTrioPC1.Open(iPortType, iPortID);

            if (axTrioPC1.IsOpen(iPortID))
            {
                btnClose.Enabled = true;
                btnRefresh.Enabled = true;
                m_iPortID = iPortID;
                axTrioPC1.Refresh();
                RefreshVisibleTab();
            }
            else
            {
                btnOpen.Enabled = true;
                m_iPortID = 0;
            }
        }

        private void cbConnectionType_SelectedIndexChanged(object sender, EventArgs e)
        {
            switch (cbConnectionType.Text)
            {
            case "USB":
                tbHostAddress.Enabled = false;
                tbChannel.Enabled = false;
                break;
            case "Ethernet":
                tbHostAddress.Enabled = true;
                tbChannel.Enabled = false;
                break;
            case "PCI":
                tbHostAddress.Enabled = false;
                tbChannel.Enabled = true;
                break;
            default:
                tbHostAddress.Enabled = false;
                tbChannel.Enabled = false;
                break;
            }
        }

        private void btnRefresh_Click(object sender, EventArgs e)
        {
            RefreshVisibleTab();
            btnRefresh.Focus();
        }

        private void RefreshVisibleTab()
        {
            btnRefresh.Enabled = false;
            btnRefreshLamp.BackColor = System.Drawing.Color.Red;
            btnRefreshLamp.Refresh();

            switch (shtControllerPages.SelectedIndex)
            {
            case 0:
                RefreshStatus();
                break;
            case 1:
                RefreshAxisParameters();
                break;
            case 2:
                RefreshIO();
                break;
            case 3:
                RefreshTable();
                break;
            case 4:
                RefreshVRs();
                break;
            case 5:
                RefreshDirectory();
                break;
            }

            btnRefreshLamp.BackColor = System.Drawing.SystemColors.Control;
            btnRefreshLamp.Refresh();
            btnRefresh.Enabled = true;
        }

        private void RefreshStatus()
        {
            if (axTrioPC1.IsOpen(m_iPortID))
            {
                double dReadVal;

                if (axTrioPC1.GetVariable("CONTROL", out dReadVal))
                    tbControllerType.Text = dReadVal.ToString();

                if (axTrioPC1.GetVariable("VERSION", out dReadVal))
                    tbSystemVersion.Text = Math.Round(dReadVal, 4).ToString();

                if (axTrioPC1.GetVariable("NIO", out dReadVal))
                    tbDigitalIOs.Text = dReadVal.ToString();

                if (axTrioPC1.GetVariable("NAIO", out dReadVal))
                    tbAnalogueIOs.Text = dReadVal.ToString();
            }
        }

        private void RefreshAxisParameters()
        {
            if (axTrioPC1.IsOpen(m_iPortID))
            {
                double dReadVar;
                short iBase = Convert.ToInt16(tbAxis.Text);
                int nItem = 0;

                if (axTrioPC1.Base(1, iBase))
                {
                    if (axTrioPC1.GetVariable("ATYPE", out dReadVar))
                        lstParameters1.Items[1].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("P_GAIN", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("I_GAIN", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("D_GAIN", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("OV_GAIN", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("VFF_GAIN", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("UNITS", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("SPEED", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("ACCEL", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("DECEL", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("CREEP", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("JOGSPEED", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("FE_LIMIT", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("DAC", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("SERVO", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("REP_DIST", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("FWD_IN", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("REV_IN", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("DATUM_IN", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("FH_IN", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("FS_LIMIT", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("RS_LIMIT", out dReadVar))
                        lstParameters1.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    nItem = 0;

                    if (axTrioPC1.GetVariable("MTYPE", out dReadVar))
                        lstParameters2.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("NTYPE", out dReadVar))
                        lstParameters2.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("MPOS", out dReadVar))
                        lstParameters2.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("DPOS", out dReadVar))
                        lstParameters2.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("FE", out dReadVar))
                        lstParameters2.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("AXISSTATUS", out dReadVar))
                        lstParameters2.Items[nItem++].SubItems[1].Text = dReadVar.ToString();

                    if (axTrioPC1.GetVariable("VPSPEED", out dReadVar))
                        lstParameters2.Items[nItem++].SubItems[1].Text = dReadVar.ToString();
                } // if ( axTrioPC1.Base( 1, iBase ))
            } // if ( axTrioPC1.IsOpen( iPortID ))
        }

        private void RefreshIO()
        {
            if (axTrioPC1.IsOpen(m_iPortID))
            {
                double dReadVar;

                if (axTrioPC1.GetVariable("NIO", out dReadVar))
                {
                    const int kiIOBlockSize = 16;
                    int iLastToGet = (int)dReadVar - 1;
                    short iFrom = 0,
                        iTo = -1;
                    int iReadVar;
                    bool bOK = true;
                    string strIn = "";
                    int nBit;

                    while (iFrom <= iLastToGet && bOK)
                    {
                        iTo = (short)(iFrom + kiIOBlockSize - 1);
                        if (iTo > iLastToGet)
                            iTo = (short)iLastToGet;

                        if (axTrioPC1.In(iFrom, iTo, out iReadVar))
                        {
                            for (nBit = 0; nBit < iTo - iFrom + 1; nBit++)
                            {
                                if ((iReadVar & 1) == 1)
                                    strIn = "1" + strIn;
                                else
                                    strIn = "0" + strIn;
                                iReadVar >>= 1;
                            }
                            iFrom = (short)(iTo + 1);
                        }
                        else
                            bOK = false;
                    }
                    if (bOK)
                        tbDigitalIOValue.Text = strIn;
                    else
                        tbDigitalIOValue.Text = "";
                } // if ( axTrioPC1.GetVariable( "NIO", out dReadVar ))
            } // if ( axTrioPC1.IsOpen( m_iPortID ))
        }

        private void RefreshTable()
        {
            if (axTrioPC1.IsOpen(m_iPortID))
            {
                int iStart = Convert.ToInt32(tbTableStart.Text),
                    iEntries = Convert.ToInt32(tbTableEntries.Text);

                lstTableValues.Items.Clear();

                if (iStart >= 0 && iEntries > 0)
                {
                    int iFrom = iStart,
                        iTo,
                        iEntry;

                    double[] dValues = new double[] { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                                                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
                    object oValues = dValues;
                    bool bOK = true;
                    ListViewItem Item;

                    while (iFrom < (iStart + iEntries) && bOK)
                    {
                        iTo = iFrom + dValues.GetLength(0) - 1;
                        if (iTo >= (iStart + iEntries))
                            iTo = iStart + iEntries - 1;
                        if (axTrioPC1.GetTable(iFrom, iTo - iFrom + 1, ref oValues))
                        {
                            dValues = (double[])oValues;
                            for (iEntry = 0; iEntry < iTo - iFrom + 1; iEntry++)
                            {
                                Item = lstTableValues.Items.Add((iFrom + iEntry).ToString());
                                Item.SubItems.Add(dValues[iEntry].ToString());
                            }
                            iFrom = iTo + 1;
                        }
                        else
                            bOK = false;
                    }
                }
            } // if ( axTrioPC1.IsOpen( m_iPortID ) )
        }

        private void RefreshVRs()
        {
            if (axTrioPC1.IsOpen(m_iPortID))
            {
                short iStart = Convert.ToInt16(tbVRStart.Text),
                    iEntries = Convert.ToInt16(tbVREntries.Text);

                if (iStart >= 0 && iEntries > 0)
                {
                    short iEntry;
                    double dReadVar;
                    bool bOK = true;
                    ListViewItem Item;

                    lstVRValues.Items.Clear();
                    for (iEntry = 0; iEntry < iEntries && bOK; iEntry++)
                    {
                        if (axTrioPC1.GetVr((short)(iEntry + iStart), out dReadVar))
                        {
                            Item = lstVRValues.Items.Add((iEntry + iStart).ToString());
                            Item.SubItems.Add(dReadVar.ToString());
                        }
                        else
                            bOK = false;
                    }
                } // if ( iStart >= 0 && iEntries > 0 )
            } // if ( axTrioPC1.IsOpen( m_iPortID ) )
        }

        private void InitParameterGrids()
        {
            int nItem;

            lstParameters1.Items.Add("ATYPE");
            lstParameters1.Items.Add("P_GAIN");
            lstParameters1.Items.Add("I_GAIN");
            lstParameters1.Items.Add("D_GAIN");
            lstParameters1.Items.Add("OV_GAIN");
            lstParameters1.Items.Add("VFF_GAIN");
            lstParameters1.Items.Add("UNITS");
            lstParameters1.Items.Add("SPEED");
            lstParameters1.Items.Add("ACCEL");
            lstParameters1.Items.Add("DECEL");
            lstParameters1.Items.Add("CREEP");
            lstParameters1.Items.Add("JOGSPEED");
            lstParameters1.Items.Add("FE_LIMIT");
            lstParameters1.Items.Add("DAC");
            lstParameters1.Items.Add("SERVO");
            lstParameters1.Items.Add("REP_DIST");
            lstParameters1.Items.Add("FWD_IN");
            lstParameters1.Items.Add("REV_IN");
            lstParameters1.Items.Add("DATUM_IN");
            lstParameters1.Items.Add("FH_IN");
            lstParameters1.Items.Add("FS_LIMIT");
            lstParameters1.Items.Add("RS_LIMIT");

            for (nItem = 0; nItem < lstParameters1.Items.Count; nItem++)
                lstParameters1.Items[nItem].SubItems.Add("0 ");

            lstParameters2.Items.Add("MTYPE");
            lstParameters2.Items.Add("NTYPE");
            lstParameters2.Items.Add("MPOS");
            lstParameters2.Items.Add("DPOS");
            lstParameters2.Items.Add("FE");
            lstParameters2.Items.Add("AXISSTATUS");
            lstParameters2.Items.Add("VPSPEED");

            for (nItem = 0; nItem < lstParameters2.Items.Count; nItem++)
                lstParameters2.Items[nItem].SubItems.Add("0 ");

            lstMoveParams.Items.Add("MTYPE");
            lstMoveParams.Items.Add("NTYPE");
            lstMoveParams.Items.Add("MPOS");
            lstMoveParams.Items.Add("DPOS");
            lstMoveParams.Items.Add("FE");
            lstMoveParams.Items.Add("SERVO");
            lstMoveParams.Items.Add("REMAIN");

            for (nItem = 0; nItem < lstMoveParams.Items.Count; nItem++)
            {
                lstMoveParams.Items[nItem].SubItems.Add("0 ");
                lstMoveParams.Items[nItem].SubItems.Add("0 ");
            }

        } // private void InitParameterGrids()

        private void tbAxis_TextChanged(object sender, System.EventArgs e)
        {
            ClearParameterGrids();
        }

        private void ClearParameterGrids()
        {
            int nItem;

            for (nItem = 0; nItem < lstParameters1.Items.Count; nItem++)
                lstParameters1.Items[nItem].SubItems[1].Text = "0";

            for (nItem = 0; nItem < lstParameters2.Items.Count; nItem++)
                lstParameters2.Items[nItem].SubItems[1].Text = "0";
        }

        private void shtControllerPages_SelectedIndexChanged(object sender, System.EventArgs e)
        {
            RefreshVisibleTab();
        }

        private void RefreshDirectory()
        {
            if (axTrioPC1.IsOpen(m_iPortID))
            {
                string strDir = "";

                if (axTrioPC1.Dir(out strDir))
                    tbDirList.Text = strDir;
            }
        }

        private void btnSetVR_Click(object sender, System.EventArgs e)
        {
            frmSetVR SVForm = new frmSetVR();

            if (SVForm.ShowDialog(this) == DialogResult.OK)
            {
                short nVR = Convert.ToInt16(SVForm.VR);

                if (nVR >= 0)
                {
                    short nStart = Convert.ToInt16(tbVRStart.Text);

                    axTrioPC1.SetVr(nVR, SVForm.Value);
                    if (nVR >= nStart && nVR < nStart + Convert.ToInt16(tbVREntries.Text))
                    {
                        RefreshVRs();
                    }
                }
            }
        }

        private void btnRun_Click(object sender, System.EventArgs e)
        {
            btnRun.Enabled = false;
            RunCommandSequence();
            btnRun.Enabled = true;
        }

        private void RunCommandSequence()
        {
            int nAxis;
            object oBases;
            object oValues;
            int[] nBases = new int[2];
            double[] dValues = new double[2];
            const double kdBaseLength = 50;

            tbRunStatus.Text = "Running";
            tbRunStatus.Refresh();

            for (nAxis = 0; nAxis <= 1; nAxis++)
            {
                axTrioPC1.Base(1, nAxis);
                axTrioPC1.SetVariable("UNITS", 1000);
                axTrioPC1.SetVariable("SPEED", 25);
                axTrioPC1.SetVariable("ACCEL", 50);
                axTrioPC1.SetVariable("DECEL", 25);
                axTrioPC1.SetVariable("FE_LIMIT", 2);
            }

            axTrioPC1.SetVariable("WDOG", 1);

            for (nAxis = 0; nAxis <= 1; nAxis++)
            {
                axTrioPC1.Base(1, nAxis);
                axTrioPC1.SetVariable("SERVO", 1);
            }

            nBases[0] = 0;
            nBases[1] = 1;
            oBases = nBases;

            axTrioPC1.Base(2, oBases);
            dValues[0] = 0.0;
            dValues[1] = 0.0;
            oValues = dValues;
            axTrioPC1.MoveAbs(2, dValues, oBases);
            WaitNextFree();

            axTrioPC1.Base(2, oBases);
            dValues[0] = kdBaseLength;
            dValues[1] = 0.0;
            oValues = dValues;
            axTrioPC1.MoveRel(2, dValues, oBases);
            WaitNextFree();

            axTrioPC1.Base(2, oBases);
            dValues[0] = -kdBaseLength;
            dValues[1] = -kdBaseLength;
            oValues = dValues;
            axTrioPC1.MoveRel(2, dValues, oBases);
            WaitNextFree();

            axTrioPC1.Base(2, oBases);
            dValues[0] = -kdBaseLength;
            dValues[1] = kdBaseLength;
            oValues = dValues;
            axTrioPC1.MoveRel(2, oValues, oBases);
            WaitNextFree();

            axTrioPC1.Base(2, oBases);
            dValues[0] = kdBaseLength;
            dValues[1] = kdBaseLength;
            oValues = dValues;
            axTrioPC1.MoveRel(2, oValues, oBases);
            WaitNextFree();

            axTrioPC1.Base(2, oBases);
            dValues[0] = kdBaseLength;
            dValues[1] = -kdBaseLength;
            oValues = dValues;
            axTrioPC1.MoveRel(2, oValues, oBases);
            WaitNextFree();

            axTrioPC1.Base(2, oBases);
            dValues[0] = 0.0;
            dValues[1] = 0.0;
            oValues = dValues;
            axTrioPC1.MoveAbs(2, oValues, oBases);
            WaitForEndOfMove();

            for (nAxis = 0; nAxis <= 1; nAxis++)
            {
                axTrioPC1.Base(1, nAxis);
                axTrioPC1.SetVariable("SERVO", 0);
            }
            axTrioPC1.SetVariable("WDOG", 0);

            GetMoveData();

            tbRunStatus.Text = "Stopped";
            tbRunStatus.Refresh();

        } // private void RunCommandSequence()


        private void WaitForEndOfMove()
        {
            double[] dRemain = new double[2];
            int nAxis;
            bool bWaiting;

            bWaiting = true;
            while (bWaiting)
            {
                for (nAxis = 0; nAxis <= 1; nAxis++)
                {
                    if (axTrioPC1.Base(1, nAxis))
                        axTrioPC1.GetVariable("REMAIN", out dRemain[nAxis]);
                    bWaiting = (Math.Abs(dRemain[0]) > 1.0) || (Math.Abs(dRemain[1]) > 1.0);
                }
                GetMoveData();
            }
        } // private void WaitForEndOfMove()

        private void WaitNextFree()
        {
            double[] dNType = new double[2];
            int nAxis;
            bool bWaiting;

            bWaiting = true;
            while (bWaiting)
            {
                for (nAxis = 0; nAxis <= 1; nAxis++)
                {
                    if (axTrioPC1.Base(1, nAxis))
                        axTrioPC1.GetVariable("NTYPE", out dNType[nAxis]);
                }
                bWaiting = (Math.Abs(dNType[0]) > 0.5) || (Math.Abs(dNType[1]) > 0.5);
                GetMoveData();
            }
        } // private void TfrmMain.WaitNextFree()

        private void GetMoveData()
        {
            int nAxis;
            int nItem;
            double dReadVar;

            SetRefreshActive(true);
            if (axTrioPC1.GetVariable("WDOG", out dReadVar))
            {
                if (dReadVar > 0.5)
                {
                    tbWatchdogState.Text = "On";
                    tbWatchdogState.BackColor = System.Drawing.Color.Red;
                }
                else
                {
                    tbWatchdogState.Text = "Off";
                    tbWatchdogState.BackColor = System.Drawing.SystemColors.Control;
                }
            }
            tbWatchdogState.Refresh();

            for (nAxis = 0; nAxis <= 1; nAxis++)
            {
                nItem = 0;

                if (axTrioPC1.Base(1, nAxis))
                {
                    if (axTrioPC1.GetVariable("MTYPE", out dReadVar))
                    {
                        lstMoveParams.Items[nItem++].SubItems[nAxis + 1].Text = dReadVar.ToString();
                        if (axTrioPC1.GetVariable("NTYPE", out dReadVar))
                        {
                            lstMoveParams.Items[nItem++].SubItems[nAxis + 1].Text = dReadVar.ToString();
                            if (axTrioPC1.GetVariable("MPOS", out dReadVar))
                            {
                                lstMoveParams.Items[nItem++].SubItems[nAxis + 1].Text = dReadVar.ToString();
                                if (axTrioPC1.GetVariable("DPOS", out dReadVar))
                                {
                                    lstMoveParams.Items[nItem++].SubItems[nAxis + 1].Text = dReadVar.ToString();
                                    if (axTrioPC1.GetVariable("FE", out dReadVar))
                                    {
                                        lstMoveParams.Items[nItem++].SubItems[nAxis + 1].Text = dReadVar.ToString();
                                        if (axTrioPC1.GetVariable("SERVO", out dReadVar))
                                        {
                                            lstMoveParams.Items[nItem++].SubItems[nAxis + 1].Text = dReadVar.ToString();
                                            if (axTrioPC1.GetVariable("REMAIN", out dReadVar))
                                                lstMoveParams.Items[nItem++].SubItems[nAxis + 1].Text = dReadVar.ToString();
                                        }
                                    }
                                }
                            }
                        }
                    }
                } // if (axTrioPC1.Base(1, nAxis))
            } // for (nAxis = 0; nAxis <= 1; nAxis++)
            lstMoveParams.Refresh();

            SetRefreshActive(false);
        } // private void GetMoveData()

        private void SetRefreshActive(bool bActive)
        {
            if (bActive)
            {
                Cursor = Cursors.WaitCursor;
                btnRefresh.Enabled = false;
                btnRefreshLamp.BackColor = System.Drawing.Color.Red;
            }
            else
            {
                Cursor = Cursors.Default;
                btnRefreshLamp.BackColor = System.Drawing.SystemColors.Control;
                btnRefresh.Enabled = true;
            }

            btnRefreshLamp.Refresh();
        } // private void SetRefreshActive(bool bActive)

        private void btnWriteTableValues_Click(object sender, System.EventArgs e)
        {
            WriteTableValues();
        } // private void SetRefreshActive(bool bActive)

        private void WriteTableValues()
        {
            const int c_nMaxAtOnce = 8;
            const double c_dMaxValue = 1e12;
            const double c_dMinValue = -1e12;

            int nStart;
            int nEntries;
            int nEntry;
            int nToSend;
            int nStartIndex;
            int nVarIndex;
            double dValue;
            object oValues;
            bool bOK = true;
            double[] dValues = new double[c_nMaxAtOnce];
            string strValue;

            SetRefreshActive(true);
            nStart = Convert.ToInt32(tbTableStart.Text);
            if (nStart < 0)
                nStart = 0;
            nEntries = Convert.ToInt32(tbTableEntries.Text);
            if (nEntries < 0)
                nEntries = 0;

            nToSend = nEntries;
            nEntry = 0;

            nStartIndex = Convert.ToInt32(lstTableValues.Items[0].SubItems[0].Text);
            while (nToSend > 0 && bOK)
            {
                // Gather group of values to send
                nVarIndex = 0;
                while ((nToSend > 0) && (nVarIndex < c_nMaxAtOnce) && bOK)
                {
                    strValue = lstTableValues.Items[nEntry].SubItems[1].Text;
                    dValue = Convert.ToDouble(strValue);
                    if (dValue > c_dMaxValue)
                        dValue = c_dMaxValue;
                    else if (dValue < c_dMinValue)
                        dValue = c_dMinValue;

                    dValues[nVarIndex] = dValue;
                    ++nVarIndex;
                    ++nEntry;
                    --nToSend;
                }
                if (nVarIndex > 0)
                {
                    oValues = dValues;
                    axTrioPC1.SetTable(nStartIndex, nVarIndex, oValues);
                    nStartIndex = nStartIndex + nVarIndex;
                }
            } // while (nToSend > 0 && bOK)
            SetRefreshActive(false);
        }

        private int LocateTableEntry(int iX, int iY)
        {
            int nItem = -1;
            bool bFound = false;
            Rectangle rectItem;

            if (iX >= 0 && iY >= 0)
            {
                // Locate Item
                nItem = 0;
                while (nItem < lstTableValues.Items.Count && !bFound)
                {
                    rectItem = lstTableValues.Items[nItem].GetBounds(ItemBoundsPortion.Label);
                    if (iY >= rectItem.Top && iY <= rectItem.Bottom)
                        bFound = true;
                    else
                        nItem++;
                }

                if (bFound)
                {
                    int nLeft = lstTableValues.Columns[0].Width;
                    int nRight = nLeft + lstTableValues.Columns[1].Width;
                    // Item only valid if column = 1
                    if (iX < nLeft || iX >= nRight)
                        nItem = -1;
                }
            } // if (iX >= 0 && iY >= 0)

            return nItem;
        }

        private void lstTableValues_MouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            m_iPreSelectedTableItem = LocateTableEntry(e.X, e.Y);
        }

        private void lstTableValues_MouseUp(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            if (m_iPreSelectedTableItem >= 0)
            {
                m_iSelectedTableItem = LocateTableEntry(e.X, e.Y);
                if (m_iPreSelectedTableItem == m_iSelectedTableItem)
                {
                    Rectangle rectValue;
                    Rectangle rectTable;

                    rectValue = lstTableValues.GetItemRect(m_iSelectedTableItem, ItemBoundsPortion.Entire);
                    rectValue.X = lstTableValues.Columns[0].Width;
                    rectValue.Width = lstTableValues.Columns[1].Width;
                    rectValue.Height -= 4;

                    rectTable = lstTableValues.Bounds;

                    rectValue.Offset(rectTable.Left + 2, rectTable.Top);
                    tbTableEdit.Bounds = rectValue;
                    tbTableEdit.Text = lstTableValues.Items[m_iSelectedTableItem].SubItems[1].Text;
                    tbTableEdit.Show();
                    tbTableEdit.Focus();
                    tbTableEdit.SelectAll();
                }
            }
        }

        private void tbTableEdit_Leave(object sender, System.EventArgs e)
        {
            lstTableValues.Items[m_iSelectedTableItem].SubItems[1].Text = tbTableEdit.Text;
            tbTableEdit.Hide();
        }

        private void frmMain_Load(object sender, EventArgs e)
        {
            InitParameterGrids();
        }
    }
}
